/**
	@file matrix.h
	@author   Thomas Stalter
	@version  1.00.00

 	@brief prototypes of functions to calculate matrices and vectors

 	Home of this file is <http://www.protokollkonverter.de>.

	@copyright Copyright: (C) 2018 Thomas Stalter

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <https://www.gnu.org/licenses/>.

	<h3>change log</h3>
	<table>
		<tr><th>date</th><th>version</th><th>author</th><th>comment</th></tr>
		<tr>
			<td>04/28/2018</td><td>1.00.00</td><td>Thomas Stalter</td><td>first release</td>
		</tr>
		<tr>
			<td>03/29/2018</td><td>0.99.00</td><td>Thomas Stalter</td><td>initial version</td>
		</tr>
	</table>
 */


#include <stddef.h>
#include <stdint.h>
#include <stdbool.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
	@defgroup matrix matrix operations
	@{
**/

#define VECTOR_STANDARD_DIMENSION	3  //!< standard dimension for vectors, normal coordination system

/**
	@typedef MATRIX
	@brief instance pointer of a matrix
**/
typedef struct _MATRIX *MATRIX;

/**
	@typedef VECTOR
	@brief instance pointer of a vector

	The pointer is the same as @ref MATRIX. It's for good programming style.
**/
typedef struct _MATRIX *VECTOR;

/**
	@typedef MATRIX_VALUE
	@brief type of supported value

	The type declares type of value to calculate with.
**/
typedef double MATRIX_VALUE;

/**
	@brief print matrix

	Function prints matrix to stdout by puts and printf.

	For example
	\f$\begin{pmatrix} 1 & 0 & 0 \\ 0 & 1 & 0 \\ 0 & 0 & 1 \end{pmatrix}\f$
	will be printed as

	@code
	+-
	|    1.000    0.000    0.000
	|    0.000    1.000    0.000
	|    0.000    0.000    1.000
	+-
	@endcode

	@param [in] A	matrix to print
**/
extern void matrixPrint(MATRIX A);

/**
	@brief create a new matrix

	Function allocates memory for a new matrix of given dimensions.
	All elements are set to 0.

	@see matrixDelete

	@param [in] rows	rows of matrix A to create
	@param [in] columns	columns of matrix A to create
	@return new matrix<br>NULL on memory allocation error
**/
extern MATRIX matrixNew(const size_t rows, const size_t columns);

/**
	@brief create a new unit matrix

	Function allocates memory for a new matrix of and initialize values.

	For example <b>matrixNewUnit(3)</b> will create a matrix like
	\f[
		\begin{pmatrix} 1 & 0 & 0 \\ 0 & 1 & 0 \\ 0 & 0 & 1 \end{pmatrix}
	\f]

	@see matrixDelete

	@param [in] dimension	dimension of rows and columns of matrix A to create
	@return new matrix<br>NULL on parameter or memory allocation error
**/
extern MATRIX matrixNewUnit(const size_t dimension);

/**
	@brief free memory allocated by @ref matrixNew

	Function frees memory of matrix created by @ref matrixNew or @ref matrixNewUnit.
	The matrix pointer A will be set to NULL.

	@param [in,out] A	pointer to matrix do delete
**/
extern void matrixDelete(MATRIX *A);

/**
	@brief get dimensions of matrix

	Function inserts dimensions of matrix in rows an columns if pointers are different to NULL.

	@param [out] rows address of variable rows to insert count of rows of matrix, can be NULL if only columns are requested
	@param [out] columns address of variable columns to insert count of columns of matrix, can be NULL if only rows are requested
	@return true on success<br>false on parameter error (A is NULL)
**/
extern bool matrixGetDimension(MATRIX A, size_t *rows, size_t *columns);

/**
	@brief compare two matrices

	Function compares matrices. Matrices are equal if they have the same dimensions and all elements are equal.

	@param [in]	A matrix to compare with matrix B
	@param [in] B matrix to compare with matrix A
	@return true if matrices are equal<br>false if matrices are not equal or parameter error
**/
extern bool matrixEquals(MATRIX A, MATRIX B);

/**
	@brief set one element in matrix

	Function sets the value of one element in matrix. Rows are count from 0..(rows of matrix - 1) and columns from 0..(columns of matrix - 1).

	@param [in]	A	matrix to set value in
	@param [in] row	row of element
	@param [in] column	column of element
	@param [in] value	value to set to element
	@return true on success<br>false on parameter error
**/
extern bool matrixSetElement(MATRIX A, const size_t row, const size_t column, const MATRIX_VALUE value);

/**
	@brief set one row in matrix

	Function sets values of one row in matrix. Rows are count from 0..(rows of matrix - 1).
	valuesCount must be equal to columns of matrix.

	@param [in]	A	matrix to set value in
	@param [in] row	row to set
	@param [in] values	values to set
	@param [in] valuesCount	count of values to set
	@return true on success<br>false on parameter error
**/
extern bool matrixSetRow(MATRIX A, const size_t row, const MATRIX_VALUE values[], const size_t valuesCount);

/**
	@brief set one column in matrix

	Function sets values of one column in matrix. Columns are count from 0..(columns of matrix - 1).

	@note valuesCount must be equal to rows of matrix.

	@param [in]	A	matrix to set value in
	@param [in] column	column to set
	@param [in] values	values to set
	@param [in] valuesCount	count of values to set
	@return true on success<br>false on parameter error
**/
extern bool matrixSetColumn(MATRIX A, const size_t column, const MATRIX_VALUE values[], const size_t valuesCount);

/**
	@brief set values of matrix

	Function sets all values of matrix. Columns are count from 0..(columns of matrix - 1).

	@note valuesCount must be equal to rows * columns of matrix.

	values must be sorted in following way:

	\f$\lbrace a_{0,0}, a_{0,1}, .. a_{0,(columns-1)}, a_{1,0}, a_{1,1}, ..., a_{(rows-1),(columns-1)} \rbrace\f$
	.

	@param [in]	A	matrix to set value in
	@param [in] values	values to set
	@param [in] valuesCount	count of values to set
	@return true on success<br>false on parameter error
**/
extern bool matrixSet(MATRIX A, const MATRIX_VALUE values[], const size_t valuesCount);

/**
	@brief get one element in matrix

	Function gets the value of one element in matrix. Rows are count from 0..(rows of matrix - 1) and columns from 0..(columns of matrix - 1).

	@param [in]	A	matrix to get value from
	@param [in] row	row of element
	@param [in] column	column of element
	@param [out] value	pointer to value to set value of element in
	@return true on success<br>false on parameter error
**/
extern bool matrixGetElement(MATRIX A, const size_t row, const size_t column, MATRIX_VALUE *value);

/**
	@brief get one row in matrix

	Function gets values of one row in matrix. Rows are count from 0..(rows of matrix - 1).

	@note valuesCount must be greater or equal to columns of matrix.

	@param [in]	A	matrix to get values from
	@param [in] row	row to get
	@param [out] values	pointer to values to copy value of elements to
	@param [in,out] valuesCount	count of values in array <br>contains afterwards count of copied values
	@return true on success<br>false on parameter error
**/
extern bool matrixGetRow(MATRIX A, const size_t row, MATRIX_VALUE values[], size_t *valuesCount);

/**
	@brief get one column in matrix

	Function gets values of one column in matrix. Columns are count from 0..(columns of matrix - 1).

	@note valuesCount must be greater or equal to rows of matrix.

	@param [in]	A	matrix to get values from
	@param [in] column	column to get
	@param [out] values	pointer to values to copy value of elements to
	@param [in,out] valuesCount	count of values in array <br>contains afterwards count of copied values
	@return true on success<br>false on parameter error
**/
extern bool matrixGetColumn(MATRIX A, const size_t column, MATRIX_VALUE values[], size_t *valuesCount);

/**
	@brief get values of matrix

	Function gets all values of matrix. Columns are count from 0..(columns of matrix - 1).

	@note valuesCount must be greater or equal to rows * columns of matrix.

	values are sorted in following way:

	\f$\lbrace a_{0,0}, a_{0,1}, .. a_{0,(columns-1)}, a_{1,0}, a_{1,1}, ..., a_{(rows-1),(columns-1)} \rbrace\f$
	.

	@param [in]	A	matrix to get values from
	@param [out] values	pointer to values to copy value of elements to
	@param [in,out] valuesCount	count of values in array <br>contains afterwards count of copied values
	@return true on success<br>false on parameter error
**/
extern bool matrixGet(MATRIX A, MATRIX_VALUE values[], size_t *valuesCount);

/**
	@brief addition of two matrices

	Function adds two matrices.

	@note Matrices must be of the same dimension.

	@param [in]	A	matrix 1 to add
	@param [in]	B	matrix 2 to add
	@return new matrix with added values<br>NULL on parameter or memory allocation error
**/
extern MATRIX matrixAddition(MATRIX A, MATRIX B);

/**
	@brief subtraction of two matrices

	Function subtract matrix B form A.

	@note Matrices must be of the same dimension.

	@param [in]	A	matrix to subtract from
	@param [in]	B	matrix to subtract from A
	@return new matrix with subtracted values<br>NULL on parameter or memory allocation error
**/
extern MATRIX matrixSubstraction(MATRIX A, MATRIX B);

/**
	@brief multiplication of two matrices

	Function multiplies matrix A and B.

	@note The count of columns of matrix A must be equal to count of rows of matrix B.

	@param [in]	A	matrix to multiply
	@param [in]	B	matrix to multiply to A
	@return new matrix with multiplied values<br>NULL on parameter or memory allocation error
**/
extern MATRIX matrixMultiplication(MATRIX A, MATRIX B);

/**
	@brief multiplication of matrix with scalar

	Function multiplies matrix A with scalar value.

	@param [in]	A	matrix to multiply
	@param [in]	scalar	value to multiply matrix A with
	@return new matrix with multiplied values<br>NULL on parameter or memory allocation error
**/
extern MATRIX matrixMultiplicationScalar(MATRIX A, MATRIX_VALUE scalar);

/**
	@brief transposition of matrix

	Function transpose matrix.

	@param [in]	A	matrix to transpose
	@return new matrix with transposed rows an columns<br>NULL on parameter or memory allocation error
**/
extern MATRIX matrixTransposition(MATRIX A);

/**
	@brief calculate inverse of matrix

	Function inverses matrix by recursive Leipniz.

	@param [in]	A	matrix to inverse
	@return new inversed matrix<br>NULL on det(matrix) == 0 or parameter or memory allocation error
**/
extern MATRIX matrixInverse(MATRIX A);

/**
	@brief calculate determinant of matrix

	Function calculate determinant of matrix by recursive Leipniz.

	@param [in]	A	matrix to calculate
	@param [out] determinant	pointer to value to set determinant of matrix in
	@return true on success<br>false on parameter error
**/
extern bool matrixDeterminant(MATRIX A, MATRIX_VALUE *determinant);

/**
	@brief clone a matrix

	Function creates a new matrix with the same dimensions and values like given matrix.

	@param [in]	A	matrix to clone
	@return new matrix like A<br>NULL on parameter or memory allocation error
**/
extern MATRIX matrixClone(MATRIX A);

/**
	@brief solve matrix A with result vector y

	Function solves system of linear equations by Cramer with determinants by recursive Leipniz.

	\f[
		A * \vec a = \vec y
	\f]

	@param [in]	A	matrix of a coefficients
	@param [in]	y	vector with results
	@return vector a on success<br>NULL if unsolvable or on parameter error
**/
extern VECTOR matrixSolve(MATRIX A, VECTOR y);

/**
	@}
	@defgroup vector vector operations
	@{
**/

/**
	@brief create a new column vector

	Function allocates memory for a new vector of given dimension.
	All elements are set to 0.

	@see vectorDelete

	@param [in] dimension	rows of vector to create
	@return new vector<br>NULL on memory allocation error
**/
extern VECTOR vectorNew(const size_t dimension);

/**
	@brief free memory allocated by @ref vectorNew

	Function frees memory of vector created by @ref vectorNew.
	The vector pointer v will be set to NULL.

	@param [in,out] v	pointer to vector do delete
**/
extern void vectorDelete(VECTOR *v);

/**
	@brief get dimension of vector

	Function sets dimension of vector in parameter dimension.

	@param [out] dimension address of variable dimension to insert count of rows of vector
	@return true on success<br>false on parameter error (v is NULL)
**/
extern bool vectorGetDimension(VECTOR v, size_t *dimension);

/**
	@brief compare two vectors

	Function compares dimensions and values of two vectors.

	@param [in] v	vector 1 to compare
	@param [in] w	vector 2 to compare
	@return true if vectors are equal<br>false vectors are different or parameter error
**/
extern bool vectorEquals(VECTOR v, VECTOR w);

/**
	@brief set one element in vector

	Function sets the value of one element in vector. Rows are count from 0..(dimension of vector - 1).

	@param [in]	v	vector to set value in
	@param [in] row	row of element
	@param [in] value	value to set to element
	@return true on success<br>false on parameter error
**/
extern bool vectorSetElement(VECTOR v, const size_t row, MATRIX_VALUE value);

/**
	@brief get one element in vector

	Function gets the value of one element in vector. Rows are count from 0..(dimension of vector - 1).

	@param [in]	v	vector to get value from
	@param [in] row	row of element
	@param [out] value	pointer to value to set value to element in
	@return true on success<br>false on parameter error
**/
extern bool vectorGetElement(VECTOR v, const size_t row, MATRIX_VALUE *value);

/**
	@brief set all values of vector

	Function sets all values of vectors.

	@note valuesCount must be equal to rows of rows.

	@param [in]	v	vector to set values in
	@param [in] values	values to set to element
	@param [in] valuesCount	count of values in array
	@return true on success<br>false on parameter error
**/
extern bool vectorSet(VECTOR v, const MATRIX_VALUE values[], const size_t valuesCount);

/**
	@brief get all values of vector

	Function gets all values of vectors.

	@note valuesCount must be greater or equal to rows of rows.

	@param [in]	v	vector to get values from
	@param [out] values	pointer to values to set values in
	@param [in,out] valuesCount	count of values in array <br>contains afterwards count of copied values
	@return true on success<br>false on parameter error
**/
extern bool vectorGet(VECTOR v, MATRIX_VALUE values[], size_t *valuesCount);

/**
	@brief add two vectors

	Function adds two vectors.

	@note The vectors must be from the same dimension \f$\vec y = \vec v + \vec w\f$.

	@param [in]	v	vector 1 to add
	@param [in]	w	vector 2 to add
	@return new vector with added values<br>NULL on parameter or memory allocation error
**/
extern VECTOR vectorAddition(VECTOR v, VECTOR w);

/**
	@brief subtract a vector

	Function subtract a vector from another vector \f$\vec y = \vec v - \vec w\f$.

	@note The vectors must be from the same dimension.

	@param [in]	v	vector to subtract from
	@param [in]	w	vector to subtract
	@return new vector with subtracted values<br>NULL on parameter or memory allocation error
**/
extern VECTOR vectorSubstraction(VECTOR v, VECTOR w);

/**
	@brief multiply a vector with scalar

	Function multiplies a vector with scalar value \f$\vec y = \vec v * scalar\f$.

	@param [in]	v	vector to multiply
	@param [in]	scalar	value to multiply vector with
	@return new vector with multiplied values<br>NULL on parameter or memory allocation error
**/
extern VECTOR vectorMultiplicationScalar(VECTOR v, MATRIX_VALUE scalar);

/**
	@brief dot product of two vectors

	Function multiplies two vectors to get dot product as scalar \f$product = \vec v \cdot \vec w\f$.

	@note The vectors must be from the same dimension.

	@param [in]	v	vector to multiply
	@param [in]	w	vector to multiply
	@param [out]	product	result value of dot multiplication
	@return true on success<br>false on parameter error
**/
extern bool vectorDotProduct(VECTOR v, VECTOR w, MATRIX_VALUE *product);

/**
	@brief cross product of two vectors

	Function multiplies two vectors to get cross product vector \f$\vec y = \vec v \times \vec w \f$.

	@note Both vectors must be from the dimension @ref VECTOR_STANDARD_DIMENSION.

	@param [in]	v	vector to multiply
	@param [in]	w	vector to multiply
	@return new vector with multiplied values<br>NULL on parameter or memory allocation error
**/
extern VECTOR vectorCrossProduct(VECTOR v, VECTOR w);

/**
	@brief scalar triple product of three vectors

	Function multiplies three vectors to get the scalar triple product \f$product = (\vec v \times \vec w) \cdot \vec x \f$.

	@note Both vectors must be from the dimension @ref VECTOR_STANDARD_DIMENSION.

	@note The vectors must be from the same dimension.

	@param [in]	v	vector to multiply
	@param [in]	w	vector to multiply
	@param [in]	x	vector to multiply
	@param [out]	product	result value of multiplication
	@return true on success<br>false on parameter error
**/
extern bool vectorScalarTripleProduct(VECTOR v, VECTOR w, VECTOR x, MATRIX_VALUE *product);

/**
	@brief outer product of two vectors

	Function multiplies two vectors to get outer product matrix \f$A = \vec v \otimes \vec w \f$.

	@param [in]	v	vector to multiply
	@param [in]	w	vector to multiply
	@return new matrix with multiplied values<br>NULL on parameter or memory allocation error
**/
extern MATRIX vectorOuterProduct(VECTOR v, VECTOR w);

/**
	@brief calculate length of vector

	Function calculates length of vector by
	\f$\sqrt{a_{0}^{2} + a_{1}^{2} + ... + a_{(rows-1)}^{2}}\f$
	.

	@param [in]	v	vector to multiply
	@param [in]	w	vector to multiply
	@return new matrix with multiplied values<br>NULL on parameter or memory allocation error
**/
extern bool vectorLength(VECTOR v, MATRIX_VALUE *length);

/**
	@brief clone a vector

	Function clones a vector.

	@note The vectors must be from the same dimension.

	Function creates a new vector with the same dimension and values like given vector.

	@param [in]	v	vector to clone
	@return new vector like v<br>NULL on parameter or memory allocation error
**/
extern VECTOR vectorClone(VECTOR v);

/**
	@}
**/

#ifdef __cplusplus
}
#endif
