/**
	@file matrix.c
	@author   Thomas Stalter
	@version  1.00.00

 	@brief implementation of functions to calculate matrices and vectors

 	Home of this file is <http://www.protokollkonverter.de>.

	@copyright Copyright: (C) 2018 Thomas Stalter

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <https://www.gnu.org/licenses/>.

	<h3>change log</h3>
	<table>
		<tr><th>date</th><th>version</th><th>author</th><th>comment</th></tr>
		<tr>
			<td>04/28/2018</td><td>1.00.00</td><td>Thomas Stalter</td><td>first release</td>
		</tr>
		<tr>
			<td>03/29/2018</td><td>0.99.00</td><td>Thomas Stalter</td><td>initial version</td>
		</tr>
	</table>
**/

#include <malloc.h>
#include <string.h>
#include <math.h>
#include <stdio.h>

#include "matrix.h"

/**
	@addtogroup matrix
	@{
**/

/**
	@struct _MATRIX
	@brief internal description of a matrix
**/
typedef struct _MATRIX {
	size_t rows;
	size_t columns;
	MATRIX_VALUE *values;
} TMATRIX;
/**
	@typedef TMATRIX
	@brief dummy to avoid compiler warning
**/

void matrixPrint(MATRIX A)
{
	if(A != NULL){
		puts("+-");
		for(register size_t r = 0; r < A->rows; r++){
			printf("|");
			for(register size_t c = 0; c < A->columns; c++){
				MATRIX_VALUE a;
				matrixGetElement(A, r, c, &a);
				printf(" %12.3f", a);
			}
			if(r < A->rows-1){
				printf("\n");
			}
		}
		puts("\n+-");
	}
	else {
		printf("%s: called with null pointer\n", __func__);
	}
}

MATRIX matrixNew(const size_t rows, const size_t columns)
{
	if(rows > 0 && columns > 0){
		MATRIX A = malloc(sizeof(struct _MATRIX));
		if(A != NULL){
			A->values = malloc(rows*columns*sizeof(MATRIX_VALUE));
			if(A->values != NULL){
				memset(A->values, 0x00, rows*columns*sizeof(MATRIX_VALUE));
				A->rows = rows;
				A->columns = columns;
				return A;
			}
			free(A);
		}
	}
	return NULL;
}

MATRIX matrixNewUnit(const size_t dimension)
{
	MATRIX E = matrixNew(dimension, dimension);
	if(E != NULL){
		for(register size_t i = 0; i < dimension; i++){
			matrixSetElement(E, i, i, 1);
		}
		return E;
	}
	return NULL;
}

void matrixDelete(MATRIX *A)
{
	if(A != NULL){
		if(*A != NULL){
			if((*A)->values != NULL){
				free((*A)->values);
			}
			free(*A);
		}
		*A = NULL;
	}
}

bool matrixGetDimension(MATRIX A, size_t *rows, size_t *columns)
{
	if(A != NULL){
		if(rows != NULL){
			*rows = A->rows;
		}
		if(columns != NULL){
			*columns = A->columns;
		}
		return true;
	}
	return false;
}

bool matrixEquals(MATRIX A, MATRIX B)
{
	if(A != NULL && B != NULL){
		if(A->rows == B->rows && A->columns == B->columns){
			for(register size_t r = 0; r < A->rows; r++){
				for(register size_t c = 0; c < A->columns; c++){
					MATRIX_VALUE a, b;
					matrixGetElement(A, r, c, &a);
					matrixGetElement(B, r, c, &b);
					if(a != b){
						return false;
					}
				}
			}
			return true;
		}
	}
	return false;
}

bool matrixSetElement(MATRIX A, const size_t row, const size_t column, const MATRIX_VALUE value)
{
	if(A != NULL){
		if(row < A->rows && column < A->columns){
			A->values[row * A->columns + column] = value;
			return true;
		}
	}
	return false;
}

bool matrixGetElement(MATRIX A, const size_t row, const size_t column, MATRIX_VALUE *value)
{
	if(A != NULL && value != NULL){
		if(row < A->rows && column < A->columns){
			*value = A->values[row * A->columns + column];
			return true;
		}
	}
	return false;
}

bool matrixSetRow(MATRIX A, const size_t row, const MATRIX_VALUE values[], const size_t valuesCount)
{
	if(A != NULL && values != NULL){
		if(row < A->rows && valuesCount == A->columns){
			for(register size_t c = 0; c < A->columns; c++){
				matrixSetElement(A, row, c, values[c]);
			}
			return true;
		}
	}
	return false;
}

bool matrixGetRow(MATRIX A, const size_t row, MATRIX_VALUE values[], size_t *valuesCount)
{
	if(A != NULL && values != NULL && valuesCount != NULL){
		if(row < A->rows && *valuesCount >= A->columns){
			for(register size_t c = 0; c < A->columns; c++){
				matrixGetElement(A, row, c, &values[c]);
			}
			*valuesCount = A->columns;
			return true;
		}
	}
	return false;
}

bool matrixSetColumn(MATRIX A, const size_t column, const MATRIX_VALUE values[], const size_t valuesCount)
{
	if(A != NULL && values != NULL){
		if(column < A->columns && valuesCount == A->rows){
			for(register size_t r = 0; r < A->rows; r++){
				matrixSetElement(A, r, column, values[r]);
			}
			return true;
		}
	}
	return false;
}

bool matrixGetColumn(MATRIX A, const size_t column, MATRIX_VALUE values[], size_t *valuesCount)
{
	if(A != NULL && values != NULL && valuesCount != NULL){
		if(column < A->columns && *valuesCount >= A->rows){
			for(register size_t r = 0; r < A->rows; r++){
				matrixGetElement(A, r, column, &values[r]);
			}
			*valuesCount = A->rows;
			return true;
		}
	}
	return false;
}

bool matrixSet(MATRIX A, const MATRIX_VALUE values[], const size_t valuesCount)
{
	if(A != NULL && values != NULL){
		if(valuesCount == A->rows*A->columns){
			register size_t i = 0;
			for(register size_t r = 0; r < A->rows; r++){
				for(register size_t c = 0; c < A->columns; c++){
					matrixSetElement(A, r, c, values[i++]);
				}
			}
			return true;
		}
	}
	return false;
}

bool matrixGet(MATRIX A, MATRIX_VALUE values[], size_t *valuesCount)
{
	if(A != NULL && values != NULL && valuesCount != NULL){
		if(*valuesCount >= A->rows*A->columns){
			register size_t i = 0;
			for(register size_t r = 0; r < A->rows; r++){
				for(register size_t c = 0; c < A->columns; c++){
					matrixGetElement(A, r, c, &values[i++]);
				}
			}
			*valuesCount = A->rows*A->columns;
			return true;
		}
	}
	return false;
}

MATRIX matrixAddition(MATRIX A, MATRIX B)
{
	if(A != NULL && B != NULL){
		if(A->rows == B->rows && A->columns == B->columns){
			MATRIX C = matrixNew(A->rows, A->columns);
			if(C != NULL){
				for(register size_t r = 0; r < A->rows; r++){
					for(register size_t c = 0; c < A->columns; c++){
						MATRIX_VALUE a, b;
						matrixGetElement(A, r, c, &a);
						matrixGetElement(B, r, c, &b);
						matrixSetElement(C, r, c, a+b);
					}
				}
				return C;
			}
		}
	}
	return NULL;
}

MATRIX matrixSubstraction(MATRIX A, MATRIX B)
{
	MATRIX C = matrixMultiplicationScalar(B, -1);
	if(C != NULL){
		MATRIX D = matrixAddition(A, C);
		matrixDelete(&C);
		return D;
	}
	return NULL;
}

MATRIX matrixMultiplication(MATRIX A, MATRIX B)
{
	if(A != NULL && B != NULL){
		if(A->columns == B->rows){
			MATRIX C = matrixNew(A->rows, B->columns);
			if(C != NULL){
				for(register size_t i = 0; i < A->rows; i++){
					for(register size_t j = 0; j < B->columns; j++){
						MATRIX_VALUE sum = 0;

						for(register size_t k = 0; k < A->columns; k++){
							MATRIX_VALUE a, b;
							matrixGetElement(A, i, k, &a);
							matrixGetElement(B, k, j, &b);
							sum += a*b;
						}
						matrixSetElement(C, i, j, sum);
					}
				}
				return C;
			}
		}
	}
	return NULL;
}

MATRIX matrixMultiplicationScalar(MATRIX A, MATRIX_VALUE scalar){
	if(A != NULL){
		MATRIX C = matrixNew(A->rows, A->columns);
		if(C != NULL){
			for(register size_t i = 0; i < A->rows; i++){
				for(register size_t j = 0; j < A->columns; j++){
					MATRIX_VALUE a;
					matrixGetElement(A, i, j, &a);
					matrixSetElement(C, i, j, a*scalar);
				}
			}
			return C;
		}
	}
	return NULL;
}

MATRIX matrixTransposition(MATRIX A)
{
	if(A != NULL){
		MATRIX C = matrixNew(A->columns, A->rows);
		if(C != NULL){
			for(register size_t i = 0; i < A->rows; i++){
				for(register size_t j = 0; j < A->columns; j++){
					MATRIX_VALUE a;
					matrixGetElement(A, i, j, &a);
					matrixSetElement(C, j, i, a);
				}
			}
			return C;
		}
	}
	return NULL;
}

/**
	@brief create a sub matrix

	Function creates a sub matrix of matrix A. The sub matrix has dimensions (rows of A) - 1 and (columns of A) - 1.
	The new matrix contains all elements except the elements of given row and column.

	\f[
		\begin{pmatrix} a_{0,0} & a_{0,1} & a_{0,1} \\ a_{1,0} & a_{1,1} & a_{1,2} \\ a_{2,0} & a_{2,1} & a_{2,2} \end{pmatrix} with
		\begin{Bmatrix} row := 1 \\ column := 2 \end{Bmatrix} =>
		\begin{vmatrix} a_{0,0} & a_{0,1} \\ a_{2,0} & a_{2,1} \end{vmatrix}
	\f]

	@param [in] A	matrix to create a sub matrix from
	@param [in] row		row of matrix A to ignore
	@param [in] column	column of matrix A to ignore
	@return new matrix<br>NULL on memory allocation error
**/
static MATRIX matrixSubDeterminant(MATRIX A, const size_t row, const size_t column){
	MATRIX C = matrixNew(A->rows-1, A->columns-1);
	if(C != NULL){
		register size_t i = 0;
		for(register size_t r = 0; r < A->rows; r++){
			if(r != row){
				register size_t j = 0;
				for(register size_t c = 0; c < A->columns; c++){
					if(c != column){
						MATRIX_VALUE a;
						matrixGetElement(A, r, c, &a);
						matrixSetElement(C, i, j, a);
						j++;
					}
				}
				i++;
			}
		}
		return C;
	}
	return NULL;
}

MATRIX matrixInverse(MATRIX A)
{
	if(A != NULL){
		if(A->rows == A->columns){
			MATRIX_VALUE determinant;
			if(matrixDeterminant(A, &determinant)){
				if(determinant != 0){
					MATRIX C = matrixNew(A->rows, A->rows);
					if(C != NULL){
						for(register size_t c = 0; c < A->columns; c++){
							for(register size_t r = 0; r < A->rows; r++){
								MATRIX_VALUE sd;
								MATRIX sub = matrixSubDeterminant(A, r, c);
								if(sub != NULL){
									matrixDeterminant(sub, &sd);
									matrixDelete(&sub);
									if(((c+r) & 1) == 1){
										sd = -sd;
									}
									matrixSetElement(C, c, r, sd/determinant);
								}
								else {
									matrixDelete(&C);
									return NULL;
								}
							}
						}
						return C;
					}
				}
			}
		}
	}
	return NULL;
}

bool matrixDeterminant(MATRIX A, MATRIX_VALUE *determinant)
{
	if(A != NULL && determinant != NULL){
		if(A->rows == A->columns){
			if(A->rows == 1){
				matrixGetElement(A, 0, 0, determinant);
			}
			else {
				/* Leibniz */
				*determinant = 0;
				for(register size_t c = 0; c < A->columns; c++){
					MATRIX_VALUE a, d;
					matrixGetElement(A, 0, c, &a);
					if((c & 1) == 1){
						a = -a;
					}

					MATRIX sub = matrixSubDeterminant(A, 0, c);
					if(sub != NULL){
						matrixDeterminant(sub, &d);
						matrixDelete(&sub);
						*determinant += a * d;
					}
					else {
						return false;
					}
				}
			}
			return true;
		}
	}
	return false;
}

MATRIX matrixClone(MATRIX A)
{
	if(A != NULL){
		MATRIX B = matrixNew(A->rows, A->columns);
		if(B != NULL){
			memcpy(B->values, A->values, A->rows*A->columns*sizeof(MATRIX_VALUE));
			return B;
		}
	}
	return NULL;
}

VECTOR matrixSolve(MATRIX A, VECTOR y)
{
	if(A != NULL && y != NULL){
		if(A->rows == y->rows){
			MATRIX_VALUE detA;
			if(matrixDeterminant(A, &detA)){
				if(detA != 0){
					VECTOR a = vectorNew(A->columns);
					for(register size_t c = 0; c < A->columns; c++){
						MATRIX Ax = matrixClone(A);
						matrixSetColumn(Ax, c, y->values, y->rows);

						MATRIX_VALUE detAx;
						if(matrixDeterminant(Ax, &detAx)){
							vectorSetElement(a, c, detAx/detA);
						}
						else {
							matrixDelete(&Ax);
							vectorDelete(&y);
							return NULL;
						}
					}
					return a;
				}
			}
		}
	}
	return NULL;
}

/**
	@}
	@addtogroup vector
	@{
**/

VECTOR vectorNew(const size_t dimension)
{
	return matrixNew(dimension, 1);
}

void vectorDelete(VECTOR *v)
{
	matrixDelete((MATRIX *)v);
}

bool vectorGetDimension(VECTOR v, size_t *dimension)
{
size_t	columns;

	if(matrixGetDimension(v, dimension, &columns)){
		if(columns == 1){
			return true;
		}
	}
	return false;
}

bool vectorEquals(VECTOR v, VECTOR w)
{
	if(v != NULL && w != NULL){
		if(v->rows == w->rows && v->columns == 1 && w->columns == 1){
			for(register size_t r = 0; r < v->rows; r++){
				MATRIX_VALUE a, b;
				vectorGetElement(v, r, &a);
				vectorGetElement(w, r, &b);
				if(a != b){
					return false;
				}
			}
			return true;
		}
	}
	return false;
}

bool vectorSetElement(VECTOR v, const size_t row, MATRIX_VALUE value)
{
	return matrixSetElement(v, row, 0, value);
}

bool vectorGetElement(VECTOR v, const size_t row, MATRIX_VALUE *value)
{
	return matrixGetElement(v, row, 0, value);
}

bool vectorSet(VECTOR v, const MATRIX_VALUE values[], const size_t valuesCount)
{
	return matrixSetColumn(v, 0, values, valuesCount);
}

bool vectorGet(VECTOR v, MATRIX_VALUE values[], size_t *valuesCount)
{
	return matrixGetColumn(v, 0, values, valuesCount);
}

VECTOR vectorAddition(VECTOR v, VECTOR w)
{
	return matrixAddition(v, w);
}

VECTOR vectorSubstraction(VECTOR v, VECTOR w)
{
	return matrixSubstraction(v, w);
}

VECTOR vectorMultiplicationScalar(VECTOR v, MATRIX_VALUE scalar)
{
	return matrixMultiplicationScalar(v, scalar);
}

bool vectorDotProduct(VECTOR v, VECTOR w, MATRIX_VALUE *product)
{
	if(v != NULL && w != NULL && product != NULL){
		if(v->rows == w->rows && v->columns == 1 && w->columns == 1){
			*product = 0;
			for(register size_t r = 0; r < v->rows; r++){
				MATRIX_VALUE a, b;
				vectorGetElement(v, r, &a);
				vectorGetElement(w, r, &b);
				*product += a*b;
			}
			return true;
		}
	}
	return false;
}

static const size_t index_a[VECTOR_STANDARD_DIMENSION] = {1, 2, 0};
static const size_t index_b[VECTOR_STANDARD_DIMENSION] = {2, 0, 1};

VECTOR vectorCrossProduct(VECTOR v, VECTOR w)
{
	if(v != NULL && w != NULL){
		if(v->rows == VECTOR_STANDARD_DIMENSION && v->rows == w->rows && v->columns == 1 && w->columns == 1){
			VECTOR z = vectorNew(v->rows);
			if(z != NULL){
				for(register size_t r = 0; r < VECTOR_STANDARD_DIMENSION; r++){
					MATRIX_VALUE a, b, sum = 0;

					vectorGetElement(v, index_a[r], &a);
					vectorGetElement(w, index_b[r], &b);
					sum += a*b;
					vectorGetElement(v, index_b[r], &a);
					vectorGetElement(w, index_a[r], &b);
					sum -= a*b;
					vectorSetElement(z, r, sum);
				}
				return z;
			}
		}
	}
	return NULL;
}

bool vectorScalarTripleProduct(VECTOR v, VECTOR w, VECTOR x, MATRIX_VALUE *product)
{
	if(v != NULL && w != NULL && x != NULL && product != NULL){
		if(v->rows == VECTOR_STANDARD_DIMENSION  && v->rows == w->rows && v->rows == x->rows && v->columns == 1 && w->columns == 1 && x->columns == 1){
			*product = 0;
			VECTOR y = vectorCrossProduct(v, w);
			if(y != NULL){
				bool success = vectorDotProduct(x, y, product);
				vectorDelete(&y);
				return success;
			}
		}
	}
	return false;
}

MATRIX vectorOuterProduct(VECTOR v, VECTOR w)
{
	if(v != NULL && w != NULL){
		if(v->rows == w->rows && v->columns == 1 && w->columns == 1){
			MATRIX x = matrixTransposition(w);
			if(x != NULL){
				MATRIX y = matrixMultiplication(v, x);
				matrixDelete(&x);
				return y;
			}
		}
	}
	return NULL;
}

bool vectorLength(VECTOR v, MATRIX_VALUE *length)
{
	if(v != NULL){
		if(v->columns == 1){
			*length = 0;
			for(register size_t r = 0; r < v->rows; r++){
				MATRIX_VALUE a;
				vectorGetElement(v, r, &a);
				*length += pow(a, 2.0);
			}
			*length = sqrt(*length);
			return true;
		}
	}
	return false;
}

VECTOR vectorClone(VECTOR v)
{
	return matrixClone(v);
}

/**
	@example testMatrix.cpp
*/

/**
	@}
**/
